(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global.PL = factory());
}(this, (function () { 'use strict';

/*eslint-disable*/
var logMsg = function logMsg(msg) {
    window.console.log(msg);
};
var logError = function logError(msg, err) {
    if (window.console.error) {
        window.console.error(msg, err);
    } else {
        logMsg(msg);
    }
};

// define console wrapper functions if applicable
var linkElems = document.getElementsByTagName('link');

// options keys
var optAlwaysPreserveOrder = 'AlwaysPreserveOrder';
var optCacheBust = 'CacheBust';
var optDebug = 'Debug';
var optBasePath = 'BasePath';

// stateless variables used across all $LAB instances
var rootPageDir = /^[^?#]*\//.exec(location.href)[0];
var rootURL = /^[\w\-]+\:\/\/\/?[^\/]+/.exec(rootPageDir)[0];
// const appendTo = document.head;

// feature detections (yay!)
var realPreloading = function () {
    // Adapted from: https://gist.github.com/igrigorik/a02f2359f3bc50ca7a9c
    var tokenList = document.createElement('link').relList;
    try {
        if (tokenList && tokenList.supports) {
            return tokenList.supports('preload');
        }
    } catch (err) {}

    return false;
}();
var scriptOrderedAsync = document.createElement('script').async === true;
var perfTiming = window.performance && window.performance.getEntriesByName;

// **************************************

function preloadResource(registryEntry) {
    var elem = document.createElement('link');

    elem.setAttribute('href', registryEntry.src);
    if (registryEntry.type === 'script' || registryEntry.type === 'module') {
        elem.setAttribute('as', 'script');
    }
    // TODO: handle more resource types
    elem.setAttribute('rel', 'preload');
    elem.setAttribute('data-from', 'PL');
    document.head.appendChild(elem);
    registryEntry.preloadRequested = true;

    return elem;
}

function loadResource(registryEntry, position) {
    if (registryEntry.type === 'script' || registryEntry.type === 'module') {
        var elem = document.createElement('script');

        elem.setAttribute('src', registryEntry.src);
        elem.setAttribute('data-from', 'PL');
        elem.async = false; // ensure ordered execution
        if (registryEntry.type === 'module') {
            elem.setAttribute('type', 'module');
        }
        document[position || 'head'].appendChild(elem);
        registryEntry.loadRequested = true;

        return elem;
    }
    if (registryEntry.opts) {
        // TODO
    }

    return null;
}

function throwGlobalError(err) {
    setTimeout(function globalError() {
        throw err;
    }, 0);
}

function assign(target, source) {
    for (var k in source) {
        target[k] = source[k];
    }

    return target;
}

// make resource URL absolute (canonical)
function canonicalURL(src, basePath) {
    var absoluteRegex = /^[\w\-]+:\/\/\/?/;

    // `src` protocol-relative (begins with `//` or `///`)?
    if (/^\/\/\/?/.test(src)) {
        src = location.protocol + src;
    }

    // `src` page-relative (neither absolute nor domain-relative beginning with `/`)?
    else if (!absoluteRegex.test(src) && src[0] != '/') {
            src = (basePath || '') + src;
        }

    // `src` still not absolute?
    if (!absoluteRegex.test(src)) {
        // domain-relative (begins with `/`)?
        if (src[0] == '/') {
            src = rootURL + src;
        }
        // otherwise, assume page-relative
        else {
                src = rootPageDir + src;
            }
    }

    return src;
}

// create an instance of $LAB
function createSandbox() {
    var registry = {};
    var defaults = {};

    defaults[optAlwaysPreserveOrder] = false;
    defaults[optCacheBust] = false;
    defaults[optDebug] = false;
    defaults[optBasePath] = '';

    if (perfTiming) {
        registerMarkupLinks();
    }

    // API for each initial $LAB instance (before chaining starts)
    var publicAPI = {
        setGlobalDefaults: setGlobalDefaults,
        setOptions: setOptions,
        script: script,
        wait: wait,
        sandbox: createSandbox
    };

    return publicAPI;

    // **************************************

    function registerMarkupLinks() {
        for (var i = 0; i < linkElems.length; i++) {
            (function loopScope(elem) {
                if (elem && /\bpreload\b/i.test(elem.getAttribute('rel'))) {
                    // must have the `as` attribute
                    var preloadAs = elem.getAttribute('as');

                    if (!preloadAs) return;

                    // canonicalize resource URL and look it up in the global performance table
                    var href = elem.getAttribute('href');
                    href = canonicalURL(href, document.baseURI); // TODO: fix document.baseURI here
                    var perfEntries = window.performance.getEntriesByName(href);

                    // already registered this resource?
                    if (href in registry) return;

                    // add global registry entry
                    var registryEntry = new createRegistryEntry(null, href);

                    registryEntry.preloadRequested = true;
                    registry[href] = registryEntry;

                    // resource already (pre)loaded?
                    if (perfEntries.length > 0) {
                        registryEntry.preloaded = true;
                    } else {
                        // listen for preload to complete
                        elem.addEventListener('load', function resourcePreloaded() {
                            // logMsg('markup link preloaded!',href);
                            elem.removeEventListener('load', resourcePreloaded);
                            registryEntry.preloaded = true;
                            notifyRegistryListeners(registryEntry);
                        });
                    }
                }
            })(linkElems[i]);
        }
    }

    // rollback `window.$LAB` to what it was before this file
    // was loaded, then return this current instance of $LAB

    function setGlobalDefaults(opts) {
        defaults = assign(defaults, opts);

        return publicAPI;
    }

    function setOptions() {
        return createChainInstance().setOptions.apply(null, arguments);
    }

    function script() {
        return createChainInstance().script.apply(null, arguments);
    }

    function wait() {
        return createChainInstance().wait.apply(null, arguments);
    }

    function createGroupEntry(check) {
        this.isGroup = true;
        this.resources = [];
        this.ready = false;
        this.complete = false;
        this.check = check || function () {};
    }

    function createRegistryEntry(type, src) {
        this.type = type;
        this.src = src;
        this.listeners = [];
        this.preloadRequested = false;
        this.preloaded = false;
        this.loadRequested = false;
        this.complete = false;
        this.opts = null;
    }

    function registerResource(resourceRecord) {
        // registry entry doesn't exist yet?
        if (!(resourceRecord.src in registry)) {
            registry[resourceRecord.src] = new createRegistryEntry(resourceRecord.type, resourceRecord.src);
        }

        var registryEntry = registry[resourceRecord.src];

        // still need to set the type for this entry?
        if (!registryEntry.type) {
            registryEntry.type = resourceRecord.type;
        }

        // any `resourceRecord` options to store?
        if (!registryEntry.opts) {
            registryEntry.opts = resourceRecord.opts;
        }

        // need to add `resourceRecord` as a listener for registry entry updates?
        if (!~registryEntry.listeners.indexOf(resourceRecord)) {
            registryEntry.listeners.push(resourceRecord);
        }

        return registryEntry;
    }

    function notifyRegistryListeners(registryEntry) {
        var groups = [];

        // collect all of the affected groups
        for (var i = 0; i < registryEntry.listeners.length; i++) {
            if (!~groups.indexOf(registryEntry.listeners[i].group)) {
                groups.push(registryEntry.listeners[i].group);
            }
        }

        // schedule all the group checks
        for (var _i = 0; _i < groups.length; _i++) {
            groups[_i].check();
        }
    }

    function createChainInstance() {
        var chainOpts = assign({}, defaults);
        var chain = [];
        var checkHook = void 0;

        var chainAPI = {
            setOptions: setOptions,
            script: script,
            wait: wait
        };

        return chainAPI;

        // **************************************

        function setOptions(opts) {
            chainOpts = assign(chainOpts, opts);

            return chainAPI;
        }

        function script() {
            for (var i = 0; i < arguments.length; i++) {
                addChainResource('script', arguments[i]);
            }
            scheduleChainCheck();

            return chainAPI;
        }

        function wait() {
            if (arguments.length > 0) {
                for (var i = 0; i < arguments.length; i++) {
                    addChainWait(arguments[i]);
                }
            } else {
                // placeholder wait entry /*waitEntry=*/
                addChainWait(true);
            }
            scheduleChainCheck();

            return chainAPI;
        }

        function addChainResource(resourceType, resourceRecord) {
            // need to add next group to chain?
            if (chain.length == 0 || !chain[chain.length - 1].isGroup || chain[chain.length - 1].complete || !realPreloading && !scriptOrderedAsync || // TODO: only apply these checks for scripts
            chainOpts[optAlwaysPreserveOrder] && !scriptOrderedAsync) {
                chain.push(new createGroupEntry(scheduleChainCheck));
            }

            var currentGroup = chain[chain.length - 1];
            currentGroup.complete = false;

            // format resource record and canonicalize URL
            if (typeof resourceRecord == 'string') {
                resourceRecord = {
                    type: resourceType,
                    src: canonicalURL(resourceRecord, chainOpts[optBasePath]),
                    group: currentGroup
                };
            } else {
                resourceRecord = {
                    type: resourceType,
                    src: canonicalURL(resourceRecord.src, chainOpts[optBasePath]),
                    opts: resourceRecord,
                    group: currentGroup
                };
            }

            // add resource to current group
            currentGroup.resources.push(resourceRecord);

            // add/lookup resource in the global registry
            var registryEntry = registerResource(resourceRecord);

            // need to start preloading this resource?
            if (realPreloading && !registryEntry.preloadRequested) {
                var elem = preloadResource(registryEntry);

                // listen for preload to complete
                elem.addEventListener('load', function resourcePreloaded() {
                    // logMsg('resource preloaded!',resourceRecord.src);
                    elem.removeEventListener('load', resourcePreloaded);
                    elem.parentNode.removeChild(elem);
                    registryEntry.preloaded = true;
                    notifyRegistryListeners(registryEntry);
                });
            }
        }

        function addChainWait(waitEntry) {
            // need empty group placeholder at beginning of chain?
            if (chain.length == 0) {
                var groupEntry = new createGroupEntry();
                groupEntry.ready = groupEntry.complete = true;
                chain.push(groupEntry);
            }

            chain.push({ wait: waitEntry });
        }

        function scheduleChainCheck() {
            if (checkHook == null) {
                checkHook = setTimeout(advanceChain, 0);
            }
        }

        function advanceChain() {
            checkHook = null;

            for (var i = 0; i < chain.length; i++) {
                if (!chain[i].complete) {
                    if (chain[i].isGroup) {
                        checkGroupStatus(chain[i]);

                        if (!chain[i].complete) {
                            break;
                        }
                    } else if (typeof chain[i].wait == 'function') {
                        try {
                            chain[i].wait();
                            chain[i].complete = true;
                        } catch (err) {
                            if (chainOpts[optDebug]) {
                                throwGlobalError(err);
                            } else {
                                logError(err);
                            }
                            break;
                        }
                    } else {
                        chain[i].complete = true;
                    }
                }
            }
        }

        function checkGroupStatus(group) {
            // is group in need of potential preload processing?
            if (!group.ready && !group.complete) {
                if (realPreloading) {
                    var groupReady = true;
                    for (var i = 0; i < group.resources.length; i++) {
                        if (!registry[group.resources[i].src].preloaded) {
                            groupReady = false;
                            break;
                        }
                    }
                    group.ready = groupReady;
                } else {
                    group.ready = true;
                }
            }

            if (!group.complete && group.ready) {
                var groupComplete = true;

                for (var _i2 = 0; _i2 < group.resources.length; _i2++) {
                    (function loopScope(resourceRecord) {
                        var registryEntry = registry[resourceRecord.src];

                        // resource loading not yet requested?
                        if (!registryEntry.loadRequested) {
                            var elem = loadResource(registryEntry, chainOpts['position']);

                            // listen for load to complete
                            elem.addEventListener('load', function resourceLoaded() {
                                // logMsg('resource loaded!',resourceRecord.src);
                                elem.removeEventListener('load', resourceLoaded);
                                registryEntry.complete = true;
                                notifyRegistryListeners(registryEntry);
                            });

                            groupComplete = false;
                        }
                        // resource loading already complete?
                        else if (registryEntry.complete) {
                                return;
                            } else {
                                groupComplete = false;

                                // need to register `resourceRecord` as a listener for currently
                                // loading (not yet complete) resource?
                                if (!~registryEntry.listeners.indexOf(resourceRecord)) {
                                    registryEntry.listeners.push(resourceRecord);
                                }
                            }
                    })(group.resources[_i2]);
                }

                if (groupComplete) {
                    group.complete = true;
                }
            }
        }
    }
}

// create the main instance of $LAB
var _Load = createSandbox();

function type(o, t) {
  return Object.prototype.toString.call(o) === '[object ' + t + ']';
}

function isArray(obj) {
  return type(obj, 'Array');
}

function isString(obj) {
  return type(obj, 'String');
}



function isFunction(obj) {
  return type(obj, 'Function');
}

function isObject(obj) {
  return type(obj, 'Object');
}



function distinct(arr) {
  if (isArray(arr)) {
    var loop = function loop(index) {
      if (index >= 1) {
        if (arr[index] === arr[index - 1]) {
          arr.splice(index, 1);
        }
        loop(index - 1);
      }
    };

    var len = arr.length;

    loop(len - 1);
  }

  return arr;
}

var CONFIG = {};

var Load = function Load(urls, resolve) {
    resolve = resolve || function () {};
    var styles = (isArray(urls) ? urls : [urls]).filter(function (url) {
        return url && url.type === 'css';
    }).map(function (u) {
        return u.link;
    });
    if (!CONFIG.enableRepeat) {
        styles = distinct(styles).filter(function (s) {
            return !checkExist(s);
        });
    }
    if (styles.length > 0) {
        styles.forEach(function (s) {
            Load.rss.push(s);
        });
        styles = styles.reverse();
    }
    var jss = (isArray(urls) ? urls : [urls]).filter(function (url) {
        return url && url.type === 'js';
    }).map(function (u) {
        return u.link;
    });
    if (!CONFIG.enableRepeat) {
        jss = distinct(jss).filter(function (j) {
            return !checkExist(j);
        });
    }
    if (jss.length > 0) {
        jss.forEach(function (j) {
            Load.rss.push(j);
        });
    } else if (!jss || isFunction(resolve)) {
        resolve();
    }
    if (styles.length > 0) {
        _useCSS(styles, function () {
            if (jss.length > 0) {
                _useJS(jss, resolve);
            } else if (isFunction(resolve)) {
                resolve();
            }
        });
    } else if (jss.length) {
        _useJS(jss, resolve);
    }
};

Load.config = function (_ref) {
    var _ref$path = _ref.path,
        path = _ref$path === undefined ? '' : _ref$path,
        _ref$enableRepeat = _ref.enableRepeat,
        enableRepeat = _ref$enableRepeat === undefined ? false : _ref$enableRepeat,
        _ref$timeout = _ref.timeout,
        timeout = _ref$timeout === undefined ? 60000 : _ref$timeout,
        _ref$position = _ref.position,
        position = _ref$position === undefined ? 'head' : _ref$position;

    CONFIG.path = path;
    CONFIG.enableRepeat = enableRepeat;
    CONFIG.timeout = timeout;
    CONFIG.position = position;
};

Load.rss = [];

function getCSSPath(url) {
    if (url.indexOf('http://') === 0 || url.indexOf('https://') === 0) {
        return url;
    }
    var base = '';
    var temp = url.split(':');

    if (temp.length === 1) {
        if (isString(CONFIG.path)) {
            return CONFIG.path + url;
        }
        if (isObject(CONFIG.path)) {
            base = CONFIG.path.main || '';

            return base + url;
        }
    }
    if (temp.length === 2) {
        if (isString(CONFIG.path)) {
            base = CONFIG.path;
        }
        if (isObject(CONFIG.path) && CONFIG.path[temp[0]]) {
            base = CONFIG.path[temp[0]];
        }

        return base + temp[1];
    }

    return base;
}

function getJSPathPrefix(url) {
    var base = '';
    var temp = url.split(':');

    if (temp.length === 1) {
        if (isString(CONFIG.path)) {
            return CONFIG.path;
        }
        if (isObject(CONFIG.path)) {
            return CONFIG.path.main || '';
        }
    }
    if (temp.length === 2 && temp[0] !== 'http' && temp[1] !== 'https') {
        if (isString(CONFIG.path)) {
            base = CONFIG.path;
        }
        if (isObject(CONFIG.path) && CONFIG.path[temp[0]]) {
            base = CONFIG.path[temp[0]];
        }
    }

    return base;
}

function getJSPathSuffix(url) {
    var temp = url.split(':');

    if (temp.length === 1) {
        return url;
    }
    if (temp.length === 2) {
        return temp[1];
    }

    return url;
}

function poll(node, callback) {
    if (callback.isCalled) {
        return;
    }
    var isLoaded = false;
    if (node.sheet) {
        isLoaded = true;
    }
    if (isLoaded) {
        setTimeout(function () {
            callback();
        }, 1);
    } else {
        setTimeout(function () {
            poll(node, callback);
        }, 1);
    }
}

function styleOnload(node, callback) {
    // for IE6-9 and Opera
    if (node.attachEvent) {
        node.attachEvent('onload', callback);
    } else {
        setTimeout(function () {
            poll(node, callback);
        }, 0);
    }
}

function loadCSS(_url, callback) {
    if (!_url || !isString(_url) || _url.indexOf('.css') < 0) return null;
    var start = _url.lastIndexOf('/');
    var end = _url.lastIndexOf('.css');
    var name = _url.substring(start + 1, end);
    var url = getCSSPath(_url);
    var links = document.getElementsByTagName('link');
    var link = document.createElement('link');
    var head = document.getElementsByTagName('head')[0];

    for (var i = 0; i < links.length; i++) {
        if (links[i].href.indexOf(url) > -1) {
            return null;
        }
    }

    link.type = 'text/css';
    link.rel = 'stylesheet';
    link.href = url;
    link.setAttribute('data-mod', name);
    link.setAttribute('data-from', 'PL');

    head.insertBefore(link, head.getElementsByTagName('link')[0] || null);

    return link;
}

var cssTime = 0;

function _useCSS(urls, callback) {
    if (urls) {
        var loadeds = [];
        var result = isArray(urls) ? urls : [urls];

        result.forEach(function (url) {
            if (!isFunction(callback)) {
                loadCSS(url);
            } else {
                styleOnload(loadCSS(url), function () {
                    loadeds.push(true);
                });
            }
        });
        if (isFunction(callback)) {
            cssTime = setInterval(function () {
                if (loadeds.length === result.length || cssTime >= CONFIG.timeout) {
                    clearInterval(cssTime);
                    if (callback && isFunction(callback)) {
                        callback();
                    }
                }
            }, 1);
        }
    }
}

function _useJS(urls, callback) {
    if (!urls) return;
    if (isArray(urls)) {
        if (urls.length > 1) {
            if (isString(urls[0])) {
                _Load.setOptions({ BasePath: getJSPathPrefix(urls[0]), position: CONFIG.position }).script(getJSPathSuffix(urls[0])).wait(function () {
                    _useJS(urls.slice(1), callback);
                });
            }
        } else if (urls.length === 1) {
            _Load.setOptions({ BasePath: getJSPathPrefix(urls[0]), position: CONFIG.position }).script(getJSPathSuffix(urls[0])).wait(callback);
        }
    } else if (isString(urls)) {
        _Load.setOptions({ BasePath: getJSPathPrefix(urls), position: CONFIG.position }).script(getJSPathSuffix(urls)).wait(callback);
    }
}

function checkExist(url) {
    return Load.rss.filter(function (rs) {
        return rs === url;
    }).length === 1;
}

Load.use = function (urls, callback) {
    var styles = (isArray(urls) ? urls : [urls]).filter(function (url) {
        return url && isString(url) && url.substring(url.length - 4) === '.css';
    });
    if (!CONFIG.enableRepeat) {
        styles = distinct(styles).filter(function (s) {
            return !checkExist(s);
        });
    }
    if (styles.length > 0) {
        styles.forEach(function (s) {
            Load.rss.push(s);
        });
        styles = styles.reverse();
    }
    var jss = (isArray(urls) ? urls : [urls]).filter(function (url) {
        return url && isString(url) && url.substring(url.length - 3) === '.js';
    });
    if (!CONFIG.enableRepeat) {
        jss = distinct(jss).filter(function (j) {
            return !checkExist(j);
        });
    }
    if (jss.length > 0) {
        jss.forEach(function (j) {
            Load.rss.push(j);
        });
    }
    if (styles.length > 0) {
        _useCSS(styles, function () {
            if (jss.length > 0) {
                _useJS(jss, callback);
            } else if (isFunction(callback)) {
                callback();
            }
        });
    } else if (jss.length) {
        _useJS(jss, callback);
    }
};

var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

var index = function (_ref) {
    var _ref$config = _ref.config,
        config = _ref$config === undefined ? {} : _ref$config,
        _ref$libs = _ref.libs,
        libs = _ref$libs === undefined ? [] : _ref$libs,
        _ref$styles = _ref.styles,
        styles = _ref$styles === undefined ? [] : _ref$styles,
        _ref$exs = _ref.exs,
        exs = _ref$exs === undefined ? [] : _ref$exs,
        _ref$loaded = _ref.loaded,
        loaded = _ref$loaded === undefined ? function () {} : _ref$loaded;

    Load.config(config);
    var _styles = styles.map(function (s) {
        return {
            type: 'css',
            link: s
        };
    });
    var _js = [].concat(toConsumableArray(libs), toConsumableArray(exs)).map(function (js) {
        return {
            type: 'js',
            link: js
        };
    });
    Load([].concat(toConsumableArray(_styles), toConsumableArray(_js)), loaded);
};

return index;

})));
