/*eslint-disable*/
const logMsg = (msg) => { window.console.log(msg); };
const logError = (msg, err) => {
    if (window.console.error) {
        window.console.error(msg, err);
    } else {
        logMsg(msg);
    }
};

// define console wrapper functions if applicable
const linkElems = document.getElementsByTagName('link');

// options keys
const optAlwaysPreserveOrder = 'AlwaysPreserveOrder';
const optCacheBust = 'CacheBust';
const optDebug = 'Debug';
const optBasePath = 'BasePath';

// stateless variables used across all $LAB instances
const rootPageDir = (/^[^?#]*\//).exec(location.href)[0];
const rootURL = (/^[\w\-]+\:\/\/\/?[^\/]+/).exec(rootPageDir)[0];
// const appendTo = document.head;

// feature detections (yay!)
const realPreloading = (() => {
    // Adapted from: https://gist.github.com/igrigorik/a02f2359f3bc50ca7a9c
    const tokenList = document.createElement('link').relList;
    try {
        if (tokenList && tokenList.supports) {
            return tokenList.supports('preload');
        }
    } catch (err) { }

    return false;
})();
const scriptOrderedAsync = document.createElement('script').async === true;
const perfTiming = window.performance && window.performance.getEntriesByName;

// **************************************

function preloadResource(registryEntry) {
    const elem = document.createElement('link');

    elem.setAttribute('href', registryEntry.src);
    if (registryEntry.type === 'script' || registryEntry.type === 'module') {
        elem.setAttribute('as', 'script');
    }
    // TODO: handle more resource types
    elem.setAttribute('rel', 'preload');
    elem.setAttribute('data-from', 'PL');
    document.head.appendChild(elem);
    registryEntry.preloadRequested = true;

    return elem;
}

function loadResource(registryEntry, position) {
    if (registryEntry.type === 'script' || registryEntry.type === 'module') {
        const elem = document.createElement('script');

        elem.setAttribute('src', registryEntry.src);
        elem.setAttribute('data-from', 'PL');
        elem.async = false;		// ensure ordered execution
        if (registryEntry.type === 'module') {
            elem.setAttribute('type', 'module');
        }
        document[position || 'head'].appendChild(elem);
        registryEntry.loadRequested = true;

        return elem;
    }
    if (registryEntry.opts) {
        // TODO
    }

    return null;
}

function throwGlobalError(err) {
    setTimeout(function globalError() { throw err; }, 0);
}

function assign(target, source) {
    for (const k in source) { target[k] = source[k]; }

    return target;
}

// make resource URL absolute (canonical)
function canonicalURL(src, basePath) {
    const absoluteRegex = /^[\w\-]+:\/\/\/?/;

    // `src` protocol-relative (begins with `//` or `///`)?
    if (/^\/\/\/?/.test(src)) {
        src = location.protocol + src;
    }

    // `src` page-relative (neither absolute nor domain-relative beginning with `/`)?
    else if (!absoluteRegex.test(src) && src[0] != '/') {
        src = (basePath || '') + src;
    }

    // `src` still not absolute?
    if (!absoluteRegex.test(src)) {
        // domain-relative (begins with `/`)?
        if (src[0] == '/') {
            src = rootURL + src;
        }
        // otherwise, assume page-relative
        else {
            src = rootPageDir + src;
        }
    }

    return src;
}

// create an instance of $LAB
function createSandbox() {
    const registry = {};
    let defaults = {};

    defaults[optAlwaysPreserveOrder] = false;
    defaults[optCacheBust] = false;
    defaults[optDebug] = false;
    defaults[optBasePath] = '';

    if (perfTiming) {
        registerMarkupLinks();
    }

    // API for each initial $LAB instance (before chaining starts)
    const publicAPI = {
        setGlobalDefaults,
        setOptions,
        script,
        wait,
        sandbox: createSandbox,
    };

    return publicAPI;

    // **************************************

    function registerMarkupLinks() {
        for (let i = 0; i < linkElems.length; i++) {
            (function loopScope(elem) {
                if (
                    elem &&
                    (/\bpreload\b/i).test(elem.getAttribute('rel'))
                ) {
                    // must have the `as` attribute
                    const preloadAs = elem.getAttribute('as');

                    if (!preloadAs) return;

                    // canonicalize resource URL and look it up in the global performance table
                    let href = elem.getAttribute('href');
                    href = canonicalURL(href, document.baseURI);	// TODO: fix document.baseURI here
                    const perfEntries = window.performance.getEntriesByName(href);

                    // already registered this resource?
                    if (href in registry) return;

                    // add global registry entry
                    const registryEntry = new createRegistryEntry(null, href);

                    registryEntry.preloadRequested = true;
                    registry[href] = registryEntry;

                    // resource already (pre)loaded?
                    if (perfEntries.length > 0) {
                        registryEntry.preloaded = true;
                    }
                    else {
                        // listen for preload to complete
                        elem.addEventListener('load', function resourcePreloaded() {
                            // logMsg('markup link preloaded!',href);
                            elem.removeEventListener('load', resourcePreloaded);
                            registryEntry.preloaded = true;
                            notifyRegistryListeners(registryEntry);
                        });
                    }
                }
            })(linkElems[i]);
        }
    }

    // rollback `window.$LAB` to what it was before this file
    // was loaded, then return this current instance of $LAB

    function setGlobalDefaults(opts) {
        defaults = assign(defaults, opts);

        return publicAPI;
    }

    function setOptions() {
        return createChainInstance().setOptions.apply(null, arguments);
    }

    function script() {
        return createChainInstance().script.apply(null, arguments);
    }

    function wait() {
        return createChainInstance().wait.apply(null, arguments);
    }

    function createGroupEntry(check) {
        this.isGroup = true;
        this.resources = [];
        this.ready = false;
        this.complete = false;
        this.check = check || function () { };
    }

    function createRegistryEntry(type, src) {
        this.type = type;
        this.src = src;
        this.listeners = [];
        this.preloadRequested = false;
        this.preloaded = false;
        this.loadRequested = false;
        this.complete = false;
        this.opts = null;
    }

    function registerResource(resourceRecord) {
        // registry entry doesn't exist yet?
        if (!(resourceRecord.src in registry)) {
            registry[resourceRecord.src] = new createRegistryEntry(resourceRecord.type, resourceRecord.src);
        }

        const registryEntry = registry[resourceRecord.src];

        // still need to set the type for this entry?
        if (!registryEntry.type) {
            registryEntry.type = resourceRecord.type;
        }

        // any `resourceRecord` options to store?
        if (!registryEntry.opts) {
            registryEntry.opts = resourceRecord.opts;
        }

        // need to add `resourceRecord` as a listener for registry entry updates?
        if (!~registryEntry.listeners.indexOf(resourceRecord)) {
            registryEntry.listeners.push(resourceRecord);
        }

        return registryEntry;
    }

    function notifyRegistryListeners(registryEntry) {
        const groups = [];

        // collect all of the affected groups
        for (let i = 0; i < registryEntry.listeners.length; i++) {
            if (!~groups.indexOf(registryEntry.listeners[i].group)) {
                groups.push(registryEntry.listeners[i].group);
            }
        }

        // schedule all the group checks
        for (let i = 0; i < groups.length; i++) {
            groups[i].check();
        }
    }

    function createChainInstance() {
        let chainOpts = assign({}, defaults);
        const chain = [];
        let checkHook;

        const chainAPI = {
            setOptions,
            script,
            wait,
        };

        return chainAPI;

        // **************************************

        function setOptions(opts) {
            chainOpts = assign(chainOpts, opts);

            return chainAPI;
        }

        function script() {
            for (let i = 0; i < arguments.length; i++) {
                addChainResource('script', arguments[i]);
            }
            scheduleChainCheck();

            return chainAPI;
        }

        function wait() {
            if (arguments.length > 0) {
                for (let i = 0; i < arguments.length; i++) {
                    addChainWait(arguments[i]);
                }
            }
            else {
                // placeholder wait entry /*waitEntry=*/
                addChainWait(true);
            }
            scheduleChainCheck();

            return chainAPI;
        }

        function addChainResource(resourceType, resourceRecord) {
            // need to add next group to chain?
            if (
                chain.length == 0 ||
                !chain[chain.length - 1].isGroup ||
                chain[chain.length - 1].complete ||
                (!realPreloading && !scriptOrderedAsync) ||		// TODO: only apply these checks for scripts
                (chainOpts[optAlwaysPreserveOrder] && !scriptOrderedAsync)
            ) {
                chain.push(new createGroupEntry(scheduleChainCheck));
            }

            const currentGroup = chain[chain.length - 1];
            currentGroup.complete = false;

            // format resource record and canonicalize URL
            if (typeof resourceRecord == 'string') {
                resourceRecord = {
                    type: resourceType,
                    src: canonicalURL(resourceRecord, chainOpts[optBasePath]),
                    group: currentGroup,
                };
            }
            else {
                resourceRecord = {
                    type: resourceType,
                    src: canonicalURL(resourceRecord.src, chainOpts[optBasePath]),
                    opts: resourceRecord,
                    group: currentGroup,
                };
            }

            // add resource to current group
            currentGroup.resources.push(resourceRecord);

            // add/lookup resource in the global registry
            const registryEntry = registerResource(resourceRecord);

            // need to start preloading this resource?
            if (realPreloading && !registryEntry.preloadRequested) {
                const elem = preloadResource(registryEntry);

                // listen for preload to complete
                elem.addEventListener('load', function resourcePreloaded() {
                    // logMsg('resource preloaded!',resourceRecord.src);
                    elem.removeEventListener('load', resourcePreloaded);
                    elem.parentNode.removeChild(elem);
                    registryEntry.preloaded = true;
                    notifyRegistryListeners(registryEntry);
                });
            }
        }

        function addChainWait(waitEntry) {
            // need empty group placeholder at beginning of chain?
            if (chain.length == 0) {
                let groupEntry = new createGroupEntry();
                groupEntry.ready = groupEntry.complete = true;
                chain.push(groupEntry);
            }

            chain.push({ wait: waitEntry });
        }

        function scheduleChainCheck() {
            if (checkHook == null) {
                checkHook = setTimeout(advanceChain, 0);
            }
        }

        function advanceChain() {
            checkHook = null;

            for (let i = 0; i < chain.length; i++) {
                if (!chain[i].complete) {
                    if (chain[i].isGroup) {
                        checkGroupStatus(chain[i]);

                        if (!chain[i].complete) {
                            break;
                        }
                    }
                    else if (typeof chain[i].wait == 'function') {
                        try {
                            chain[i].wait();
                            chain[i].complete = true;
                        }
                        catch (err) {
                            if (chainOpts[optDebug]) {
                                throwGlobalError(err);
                            }
                            else {
                                logError(err);
                            }
                            break;
                        }
                    }
                    else {
                        chain[i].complete = true;
                    }
                }
            }
        }

        function checkGroupStatus(group) {
            // is group in need of potential preload processing?
            if (!group.ready && !group.complete) {
                if (realPreloading) {
                    let groupReady = true;
                    for (let i = 0; i < group.resources.length; i++) {
                        if (!registry[group.resources[i].src].preloaded) {
                            groupReady = false;
                            break;
                        }
                    }
                    group.ready = groupReady;
                }
                else {
                    group.ready = true;
                }
            }

            if (!group.complete && group.ready) {
                let groupComplete = true;

                for (let i = 0; i < group.resources.length; i++) {
                    (function loopScope(resourceRecord) {
                        const registryEntry = registry[resourceRecord.src];

                        // resource loading not yet requested?
                        if (!registryEntry.loadRequested) {
                            const elem = loadResource(registryEntry, chainOpts['position']);

                            // listen for load to complete
                            elem.addEventListener('load', function resourceLoaded() {
                                // logMsg('resource loaded!',resourceRecord.src);
                                elem.removeEventListener('load', resourceLoaded);
                                registryEntry.complete = true;
                                notifyRegistryListeners(registryEntry);
                            });

                            groupComplete = false;
                        }
                        // resource loading already complete?
                        else if (registryEntry.complete) {
                            return;
                        }
                        else {
                            groupComplete = false;

                            // need to register `resourceRecord` as a listener for currently
                            // loading (not yet complete) resource?
                            if (!~registryEntry.listeners.indexOf(resourceRecord)) {
                                registryEntry.listeners.push(resourceRecord);
                            }
                        }
                    })(group.resources[i]);
                }

                if (groupComplete) {
                    group.complete = true;
                }
            }
        }
    }
}

// create the main instance of $LAB
export default createSandbox();